@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2013-2014  Kouhei Sutou <kou@clear-code.com>
+  Copyright(C) 2013-2018  Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
@@ -18,6 +18,14 @@
   MA 02110-1301, USA
 */
 
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif
+
+#ifdef GROONGA_NORMALIZER_MYSQL_EMBED
+#  define GRN_PLUGIN_FUNCTION_TAG normalizers_mysql
+#endif
+
 #include <groonga/normalizer.h>
 #include <groonga/nfkc.h>
 
@@ -27,6 +35,13 @@
 #include "mysql_general_ci_table.h"
 #include "mysql_unicode_ci_table.h"
 #include "mysql_unicode_ci_except_kana_ci_kana_with_voiced_sound_mark_table.h"
+#include "mysql_unicode_520_ci_table.h"
+#include "mysql_unicode_520_ci_except_kana_ci_kana_with_voiced_sound_mark_table.h"
+#include "mysql_unicode_900_ai_ci_table.h"
+#include "mysql_unicode_900_as_ci_table.h"
+#include "mysql_unicode_900_as_cs_table.h"
+#include "mysql_unicode_900_ja_as_cs_table.h"
+#include "mysql_unicode_900_ja_as_cs_ks_table.h"
 
 #ifdef __GNUC__
 #  define GNUC_UNUSED __attribute__((__unused__))
@@ -147,7 +162,7 @@ utf8_to_unichar(const char *utf8, int byte_size)
 
 static inline void
 decompose_character(const char *rest, int character_length,
-                    int *page, uint32_t *low_code)
+                    size_t *page, uint32_t *low_code)
 {
   switch (character_length) {
   case 1 :
@@ -167,10 +182,27 @@ decompose_character(const char *rest, int character_length,
       ((rest[0] & 0x07) << 10) +
       ((rest[1] & 0x3f) << 4) +
       ((rest[2] & 0x3c) >> 2);
-    *low_code = ((rest[1] & 0x03) << 6) + (rest[2] & 0x3f);
+    *low_code = ((rest[2] & 0x03) << 6) + (rest[3] & 0x3f);
+    break;
+  case 5 :
+    *page =
+      ((rest[0] & 0x03) << 16) +
+      ((rest[1] & 0x3f) << 10) +
+      ((rest[2] & 0x3f) << 4) +
+      ((rest[3] & 0x3c) >> 2);
+    *low_code = ((rest[3] & 0x03) << 6) + (rest[4] & 0x3f);
+    break;
+  case 6 :
+    *page =
+      ((rest[0] & 0x01) << 22) +
+      ((rest[1] & 0x3f) << 16) +
+      ((rest[2] & 0x3f) << 10) +
+      ((rest[3] & 0x3f) << 4) +
+      ((rest[4] & 0x3c) >> 2);
+    *low_code = ((rest[4] & 0x03) << 6) + (rest[5] & 0x3f);
     break;
   default :
-    *page = -1;
+    *page = (size_t)-1;
     *low_code = 0x00;
     break;
   }
@@ -179,23 +211,27 @@ decompose_character(const char *rest, int character_length,
 static inline void
 normalize_character(const char *utf8, int character_length,
                     uint32_t **normalize_table,
+                    size_t normalize_table_size,
                     char *normalized,
                     unsigned int *normalized_character_length,
                     unsigned int *normalized_length_in_bytes,
                     unsigned int *normalized_n_characters)
 {
-  int page;
+  size_t page;
   uint32_t low_code;
   decompose_character(utf8, character_length, &page, &low_code);
-  if ((0x00 <= page && page <= 0xff) && normalize_table[page]) {
+  if (page < normalize_table_size && normalize_table[page]) {
     uint32_t normalized_code;
     unsigned int n_bytes;
     normalized_code = normalize_table[page][low_code];
-    if (normalized_code != 0) {
+    if (normalized_code == 0x00000) {
+      *normalized_character_length = 0;
+    } else {
       n_bytes = unichar_to_utf8(normalized_code,
                                 normalized + *normalized_length_in_bytes);
       *normalized_character_length = n_bytes;
       *normalized_length_in_bytes += n_bytes;
+      (*normalized_n_characters)++;
     }
   } else {
     int i;
@@ -204,8 +240,8 @@ normalize_character(const char *utf8, int character_length,
     }
     *normalized_character_length = character_length;
     *normalized_length_in_bytes += character_length;
+    (*normalized_n_characters)++;
   }
-  (*normalized_n_characters)++;
 }
 
 static void
@@ -293,6 +329,7 @@ static void
 normalize(grn_ctx *ctx, grn_obj *string,
           const char *normalizer_type_label,
           uint32_t **normalize_table,
+          size_t normalize_table_size,
           normalizer_func custom_normalizer)
 {
   const char *original, *rest;
@@ -332,41 +369,51 @@ normalize(grn_ctx *ctx, grn_obj *string,
   rest_length = original_length_in_bytes;
   while (rest_length > 0) {
     int character_length;
+    grn_bool custom_normalized = GRN_FALSE;
+    unsigned int normalized_character_length;
+    unsigned int previous_normalized_length_in_bytes =
+      normalized_length_in_bytes;
+    unsigned int previous_normalized_n_characters =
+      normalized_n_characters;
 
     character_length = grn_plugin_charlen(ctx, rest, rest_length, encoding);
     if (character_length == 0) {
       break;
     }
 
-    if (remove_blank_p && character_length == 1 && rest[0] == ' ') {
+    if (custom_normalizer) {
+      custom_normalized = custom_normalizer(ctx,
+                                            rest,
+                                            &character_length,
+                                            rest_length - character_length,
+                                            normalize_table,
+                                            normalized,
+                                            &normalized_character_length,
+                                            &normalized_length_in_bytes,
+                                            &normalized_n_characters);
+    }
+    if (!custom_normalized) {
+      normalize_character(rest, character_length,
+                          normalize_table, normalize_table_size,
+                          normalized,
+                          &normalized_character_length,
+                          &normalized_length_in_bytes,
+                          &normalized_n_characters);
+    }
+
+    if (remove_blank_p &&
+        normalized_character_length == 1 &&
+        normalized[previous_normalized_length_in_bytes] == ' ') {
       if (current_type > types) {
         current_type[-1] |= GRN_CHAR_BLANK;
       }
       if (current_check) {
         current_check[0]++;
       }
+      normalized_length_in_bytes = previous_normalized_length_in_bytes;
+      normalized_n_characters = previous_normalized_n_characters;
     } else {
-      grn_bool custom_normalized = GRN_FALSE;
-      unsigned int normalized_character_length;
-      if (custom_normalizer) {
-        custom_normalized = custom_normalizer(ctx,
-                                              rest,
-                                              &character_length,
-                                              rest_length - character_length,
-                                              normalize_table,
-                                              normalized,
-                                              &normalized_character_length,
-                                              &normalized_length_in_bytes,
-                                              &normalized_n_characters);
-      }
-      if (!custom_normalized) {
-        normalize_character(rest, character_length, normalize_table,
-                            normalized,
-                            &normalized_character_length,
-                            &normalized_length_in_bytes,
-                            &normalized_n_characters);
-      }
-      if (current_type) {
+      if (current_type && normalized_character_length > 0) {
         char *current_normalized;
         current_normalized =
           normalized + normalized_length_in_bytes - normalized_character_length;
@@ -375,14 +422,16 @@ normalize(grn_ctx *ctx, grn_obj *string,
         current_type++;
       }
       if (current_check) {
-        unsigned int i;
         current_check[0] += character_length;
-        current_check++;
-        for (i = 1; i < normalized_character_length; i++) {
-          current_check[0] = 0;
+        if (normalized_character_length > 0) {
+          unsigned int i;
           current_check++;
+          for (i = 1; i < normalized_character_length; i++) {
+            current_check[0] = 0;
+            current_check++;
+          }
+          current_check[0] = 0;
         }
-        current_check[0] = 0;
       }
     }
 
@@ -435,7 +484,9 @@ mysql_general_ci_next(GNUC_UNUSED grn_ctx *ctx,
                      grn_encoding_to_string(encoding));
     return NULL;
   }
-  normalize(ctx, string, normalizer_type_label, general_ci_table, NULL);
+  normalize(ctx, string, normalizer_type_label,
+            general_ci_table, sizeof(general_ci_table) / sizeof(uint32_t *),
+            NULL);
   return NULL;
 }
 
@@ -459,7 +510,9 @@ mysql_unicode_ci_next(GNUC_UNUSED grn_ctx *ctx,
                      grn_encoding_to_string(encoding));
     return NULL;
   }
-  normalize(ctx, string, normalizer_type_label, unicode_ci_table, NULL);
+  normalize(ctx, string, normalizer_type_label,
+            unicode_ci_table, sizeof(unicode_ci_table) / sizeof(uint32_t *),
+            NULL);
   return NULL;
 }
 
@@ -625,10 +678,273 @@ mysql_unicode_ci_except_kana_ci_kana_with_voiced_sound_mark_next(
   normalize(ctx, string,
             normalizer_type_label,
             unicode_ci_except_kana_ci_kana_with_voiced_sound_mark_table,
+            sizeof(unicode_ci_except_kana_ci_kana_with_voiced_sound_mark_table) / sizeof(uint32_t *),
             normalize_halfwidth_katakana_with_voiced_sound_mark);
   return NULL;
 }
 
+static grn_obj *
+mysql_unicode_520_ci_next(GNUC_UNUSED grn_ctx *ctx,
+                          GNUC_UNUSED int nargs,
+                          grn_obj **args,
+                          GNUC_UNUSED grn_user_data *user_data)
+{
+  grn_obj *string = args[0];
+  grn_encoding encoding;
+  const char *normalizer_type_label = "mysql-unicode-520-ci";
+
+  encoding = grn_string_get_encoding(ctx, string);
+  if (encoding != GRN_ENC_UTF8) {
+    GRN_PLUGIN_ERROR(ctx,
+                     GRN_FUNCTION_NOT_IMPLEMENTED,
+                     "[normalizer][%s] "
+                     "UTF-8 encoding is only supported: %s",
+                     normalizer_type_label,
+                     grn_encoding_to_string(encoding));
+    return NULL;
+  }
+  normalize(ctx, string, normalizer_type_label,
+            unicode_520_ci_table,
+            sizeof(unicode_520_ci_table) / sizeof(uint32_t *),
+            NULL);
+  return NULL;
+}
+
+static grn_obj *
+mysql_unicode_520_ci_except_kana_ci_kana_with_voiced_sound_mark_next(
+  GNUC_UNUSED grn_ctx *ctx,
+  GNUC_UNUSED int nargs,
+  grn_obj **args,
+  GNUC_UNUSED grn_user_data *user_data)
+{
+  grn_obj *string = args[0];
+  grn_encoding encoding;
+  const char *normalizer_type_label =
+    "mysql-unicode-520-ci-except-kana-ci-kana-with-voiced-sound-mark";
+
+  encoding = grn_string_get_encoding(ctx, string);
+  if (encoding != GRN_ENC_UTF8) {
+    GRN_PLUGIN_ERROR(ctx,
+                     GRN_FUNCTION_NOT_IMPLEMENTED,
+                     "[normalizer][%s] "
+                     "UTF-8 encoding is only supported: %s",
+                     normalizer_type_label,
+                     grn_encoding_to_string(encoding));
+    return NULL;
+  }
+  normalize(ctx, string,
+            normalizer_type_label,
+            unicode_520_ci_except_kana_ci_kana_with_voiced_sound_mark_table,
+            sizeof(unicode_520_ci_except_kana_ci_kana_with_voiced_sound_mark_table) / sizeof(uint32_t *),
+            normalize_halfwidth_katakana_with_voiced_sound_mark);
+  return NULL;
+}
+
+typedef enum {
+  MYSQL_UNICODE_900_LOCALE_NONE,
+  MYSQL_UNICODE_900_LOCALE_JA,
+} mysql_unicode_900_locale;
+
+typedef struct {
+  uint8_t weight_level;
+  mysql_unicode_900_locale locale;
+} mysql_unicode_900_options;
+
+static void
+mysql_unicode_900_options_init(mysql_unicode_900_options *options)
+{
+  options->weight_level = 1;
+  options->locale = MYSQL_UNICODE_900_LOCALE_NONE;
+}
+
+static void *
+mysql_unicode_900_open_options(grn_ctx *ctx,
+                               GNUC_UNUSED grn_obj *string,
+                               grn_obj *raw_options,
+                               GNUC_UNUSED void *user_data)
+{
+  const char *normalizer_type_label = "mysql-unicode-900";
+  mysql_unicode_900_options *options;
+
+  options = GRN_PLUGIN_MALLOC(ctx, sizeof(*options));
+  if (!options) {
+    GRN_PLUGIN_ERROR(ctx,
+                     GRN_NO_MEMORY_AVAILABLE,
+                     "[normalizer][%s] "
+                     "failed to allocate memory for options",
+                     normalizer_type_label);
+    return NULL;
+  }
+
+  mysql_unicode_900_options_init(options);
+
+  GRN_OPTION_VALUES_EACH_BEGIN(ctx, raw_options, i, name, name_length) {
+    /* TODO: Use grn_raw_string. */
+#define STRING_EQUAL(raw_string, raw_string_length, c_string)   \
+    (raw_string_length == strlen(c_string) &&                   \
+     strncmp(raw_string, c_string, raw_string_length) == 0)
+    if (STRING_EQUAL(name, name_length, "weight_level")) {
+      options->weight_level =
+        grn_vector_get_element_uint8(ctx,
+                                     raw_options,
+                                     i,
+                                     options->weight_level);
+    } else if (STRING_EQUAL(name, name_length, "locale")) {
+      const char *locale;
+      unsigned int locale_length;
+      grn_id domain;
+      locale_length =
+        grn_vector_get_element(ctx, raw_options, i, &locale, NULL, &domain);
+      if (!grn_type_id_is_text_family(ctx, domain)) {
+        GRN_PLUGIN_FREE(ctx, options);
+        options = NULL;
+        {
+          grn_obj value;
+          grn_obj inspected;
+          GRN_OBJ_INIT(&value, GRN_BULK, 0, domain);
+          grn_bulk_write(ctx, &value, locale, locale_length);
+          GRN_TEXT_INIT(&inspected, 0);
+          grn_inspect(ctx, &value, &inspected);
+          GRN_OBJ_FIN(ctx, &value);
+          GRN_PLUGIN_ERROR(ctx,
+                           GRN_INVALID_ARGUMENT,
+                           "[normalizer][%s] "
+                           "locale must be text: <%.*s>",
+                           normalizer_type_label,
+                           (int)GRN_TEXT_LEN(&inspected),
+                           GRN_TEXT_VALUE(&inspected));
+          GRN_OBJ_FIN(ctx, &inspected);
+        }
+        break;
+      }
+      if (STRING_EQUAL(locale, locale_length, "none")) {
+        options->locale = MYSQL_UNICODE_900_LOCALE_NONE;
+      } else if (STRING_EQUAL(locale, locale_length, "ja")) {
+        options->locale = MYSQL_UNICODE_900_LOCALE_JA;
+      } else {
+        GRN_PLUGIN_FREE(ctx, options);
+        options = NULL;
+        GRN_PLUGIN_ERROR(ctx,
+                         GRN_INVALID_ARGUMENT,
+                         "[normalizer][%s] "
+                         "locale must be <none> or <ja>: <%.*s>",
+                         normalizer_type_label,
+                         (int)locale_length,
+                         locale);
+        break;
+      }
+    }
+#undef NAME_EQUAL
+  } GRN_OPTION_VALUES_EACH_END();
+
+  return options;
+}
+
+static void
+mysql_unicode_900_close_options(grn_ctx *ctx, void *data)
+{
+  mysql_unicode_900_options *options = data;
+  GRN_PLUGIN_FREE(ctx, options);
+}
+
+static grn_obj *
+mysql_unicode_900_next(grn_ctx *ctx,
+                       GNUC_UNUSED int n_args,
+                       grn_obj **args,
+                       GNUC_UNUSED grn_user_data *user_data)
+{
+  const char *normalizer_type_label = "mysql-unicode-900";
+  grn_obj *string = args[0];
+  grn_encoding encoding;
+  grn_obj *table;
+  mysql_unicode_900_options *options;
+  mysql_unicode_900_options options_raw;
+
+  encoding = grn_string_get_encoding(ctx, string);
+  if (encoding != GRN_ENC_UTF8) {
+    GRN_PLUGIN_ERROR(ctx,
+                     GRN_FUNCTION_NOT_IMPLEMENTED,
+                     "[normalizer][%s] "
+                     "UTF-8 encoding is only supported: %s",
+                     normalizer_type_label,
+                     grn_encoding_to_string(encoding));
+    return NULL;
+  }
+
+  table = grn_string_get_table(ctx, string);
+  if (table) {
+    options = grn_table_cache_normalizer_options(ctx,
+                                                 table,
+                                                 string,
+                                                 mysql_unicode_900_open_options,
+                                                 mysql_unicode_900_close_options,
+                                                 NULL);
+    if (ctx->rc != GRN_SUCCESS) {
+      return NULL;
+    }
+  } else {
+    mysql_unicode_900_options_init(&options_raw);
+    options = &options_raw;
+  }
+  if (options->weight_level == 1) {
+    normalize(ctx, string,
+              normalizer_type_label,
+              unicode_900_ai_ci_table,
+              sizeof(unicode_900_ai_ci_table) / sizeof(uint32_t *),
+              NULL);
+  } else if (options->weight_level == 2) {
+    normalize(ctx, string,
+              normalizer_type_label,
+              unicode_900_as_ci_table,
+              sizeof(unicode_900_as_ci_table) / sizeof(uint32_t *),
+              NULL);
+  } else if (options->weight_level == 3) {
+    switch (options->locale) {
+    case MYSQL_UNICODE_900_LOCALE_JA :
+      normalize(ctx, string,
+                normalizer_type_label,
+                unicode_900_ja_as_cs_table,
+                sizeof(unicode_900_ja_as_cs_table) / sizeof(uint32_t *),
+                NULL);
+      break;
+    default :
+      normalize(ctx, string,
+                normalizer_type_label,
+                unicode_900_as_cs_table,
+                sizeof(unicode_900_as_cs_table) / sizeof(uint32_t *),
+                NULL);
+      break;
+    }
+  } else if (options->weight_level == 4) {
+    switch (options->locale) {
+    case MYSQL_UNICODE_900_LOCALE_JA :
+      normalize(ctx, string,
+                normalizer_type_label,
+                unicode_900_ja_as_cs_ks_table,
+                sizeof(unicode_900_ja_as_cs_ks_table) / sizeof(uint32_t *),
+                NULL);
+      break;
+    default :
+      GRN_PLUGIN_ERROR(ctx,
+                       GRN_FUNCTION_NOT_IMPLEMENTED,
+                       "[normalizer][%s] "
+                       "locale must be ja for weight level 4",
+                       normalizer_type_label);
+      break;
+    }
+  } else {
+    GRN_PLUGIN_ERROR(ctx,
+                     GRN_FUNCTION_NOT_IMPLEMENTED,
+                     "[normalizer][%s] "
+                     "weight level must be 1, 2, 3 or 4: %u",
+                     normalizer_type_label,
+                     options->weight_level);
+    return NULL;
+  }
+
+  return NULL;
+}
+
 grn_rc
 GRN_PLUGIN_INIT(grn_ctx *ctx)
 {
@@ -651,6 +967,23 @@ GRN_PLUGIN_REGISTER(grn_ctx *ctx)
                           NULL,
                           mysql_unicode_ci_except_kana_ci_kana_with_voiced_sound_mark_next,
                           NULL);
+  grn_normalizer_register(ctx, "NormalizerMySQLUnicode520CI", -1,
+                          NULL, mysql_unicode_520_ci_next, NULL);
+  grn_normalizer_register(ctx,
+                          "NormalizerMySQLUnicode520CI"
+                          "Except"
+                          "KanaCI"
+                          "KanaWithVoicedSoundMark",
+                          -1,
+                          NULL,
+                          mysql_unicode_520_ci_except_kana_ci_kana_with_voiced_sound_mark_next,
+                          NULL);
+  grn_normalizer_register(ctx,
+                          "NormalizerMySQLUnicode900",
+                          -1,
+                          NULL,
+                          mysql_unicode_900_next,
+                          NULL);
   return GRN_SUCCESS;
 }
 
